#include "OrientedTactileSensorUnit.h"
#include <termios.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <math.h>
#include <ArmarXCore/core/application/Application.h>
#include <ArmarXCore/observers/variant/TimestampVariant.h>

using namespace armarx;

OrientedTactileSensorUnit::OrientedTactileSensorUnit()
{

    sampleIndexRotation = 0;
    sampleIndexPressure = 0;
    sampleIndexAcceleration = 0;
}

void OrientedTactileSensorUnit::onInitComponent()
{
    maxSamplesRotation = getProperty<std::size_t>("SamplesRotation").getValue();
    maxSamplesPressure = getProperty<std::size_t>("SamplesPressure").getValue();
    maxSamplesAcceleration = getProperty<std::size_t>("SamplesAcceleration").getValue();

    std::string topicName = getProperty<std::string>("TopicName").getValue();
    offeringTopic(topicName);

    //open serial port
    std::string portname = getProperty<std::string>("SerialInterfaceDevice").getValue();
    arduino.open(getProperty<std::string>("SerialInterfaceDevice").getValue(), std::ios::in);

    fd = open(portname.c_str(), O_RDONLY | O_NOCTTY);
    struct termios toptions;

    /* Get currently set options for the tty */
    tcgetattr(fd, &toptions);

    /* Set custom options */
    cfsetispeed(&toptions, B115200);
    cfsetospeed(&toptions, B115200);

    /* 8 bits, no parity, no stop bits */
    toptions.c_lflag = 0;
    toptions.c_iflag = 0;
    toptions.c_oflag = 0;

    /* commit the options */
    tcsetattr(fd, TCSANOW, &toptions);

    /* Wait for the Arduino to reset */
    usleep(1000 * 1000);

    /* Flush anything already in the serial buffer */
    tcflush(fd, TCIFLUSH);

    ARMARX_INFO << "opening device " << getProperty<std::string>("SerialInterfaceDevice").getValue();

    if (!arduino.is_open())
    {

        throw LocalException("Cannot open Arduino on ") << getProperty<std::string>("SerialInterfaceDevice").getValue();
    }

    ARMARX_INFO << "Arduino restarts, please wait ...";

    //wait for the Arduino to reboot
    usleep(4000000);
    std::string arduinoLine;
    //wait for the IMU to be calibrated or load calibration
    ARMARX_INFO << "waiting for IMU calibration - this can take some time";
    if (getProperty<bool>("calibrateSensor").getValue())
    {
        //calibrate

        while (arduinoLine.find("mode") == std::string::npos)
        {
            getline(arduino, arduinoLine, '\n');
        }
        arduino.close();

        arduino.open(getProperty<std::string>("SerialInterfaceDevice").getValue(), std::ios::out);
        arduino << "calibrate";
        arduino.flush();
        arduino.close();

        arduino.open(getProperty<std::string>("SerialInterfaceDevice").getValue(), std::ios::in);
        while (arduinoLine.find("Calibration Sucessfull") == std::string::npos)
        {
            getline(arduino, arduinoLine, '\n');
            ARMARX_INFO << arduinoLine;
        }
        getline(arduino, arduinoLine, '\n');
        if (getCalibrationValues(arduinoLine))
        {
            ARMARX_IMPORTANT << "calibrated sensor";
        }
    }
    else
    {
        //load calibration
        ARMARX_INFO << "load calibration data " << getProperty<std::string>("CalibrationData").getValue();
        if (loadCalibration())
        {
            ARMARX_IMPORTANT << "loaded calibration";
        }
    }
    readTask = new RunningTask<OrientedTactileSensorUnit>(this, &OrientedTactileSensorUnit::run);
    readTask->start();

}

void OrientedTactileSensorUnit::onConnectComponent()
{
    std::string topicName = getProperty<std::string>("TopicName").getValue();
    topicPrx = getTopic<OrientedTactileSensorUnitListenerPrx>(topicName);
}

PropertyDefinitionsPtr OrientedTactileSensorUnit::createPropertyDefinitions()
{
    return PropertyDefinitionsPtr(new OrientedTactileSensorUnitPropertyDefinitions(
                                      getConfigIdentifier()));
}

void OrientedTactileSensorUnit::run()
{
    while (readTask->isRunning())
    {
        std::string line;
        getline(arduino, line, '\n');
        SensorData data = getValues(line.c_str());
        IceUtil::Time now = IceUtil::Time::now();
        TimestampVariantPtr nowTimestamp = new TimestampVariant(now);

        //compute rotationRate
        float rotationRate = 0;

        //condition for inverse quaternion
        if ((pow(data.qw, 2) + pow(data.qx, 2) + pow(data.qy, 2) + pow(data.qz, 2)) != 0)
        {
            RotationRate sampleRotation;
            sampleRotation.timestamp = now;
            sampleRotation.orientation =  Eigen::Quaternionf(data.qw, data.qx, data.qy, data.qz);
            if (samplesRotation.size() < maxSamplesRotation)
            {
                samplesRotation.push_back(sampleRotation);
            }
            else
            {
                samplesRotation[sampleIndexRotation].timestamp = sampleRotation.timestamp;
                samplesRotation[sampleIndexRotation].orientation = sampleRotation.orientation;
                sampleIndexRotation = (sampleIndexRotation + 1) % maxSamplesRotation;
                RotationRate oldsampleRotation;
                oldsampleRotation.timestamp = samplesRotation.at(sampleIndexRotation).timestamp;
                oldsampleRotation.orientation = samplesRotation.at(sampleIndexRotation).orientation;
                Eigen::AngleAxisf aa(sampleRotation.orientation * oldsampleRotation.orientation.inverse());
                rotationRate = aa.angle() / (sampleRotation.timestamp - oldsampleRotation.timestamp).toSecondsDouble();
            }
        }

        //compute pressureRate
        float pressureRate = 0;
        PressureRate samplePressure;
        samplePressure.timestamp = now;
        samplePressure.pressure = data.pressure;
        if (samplesPressure.size() < maxSamplesPressure)
        {
            samplesPressure.push_back(samplePressure);
        }
        else
        {
            samplesPressure[sampleIndexPressure] = samplePressure;
            sampleIndexPressure = (sampleIndexPressure + 1) % maxSamplesPressure;
            PressureRate oldsamplePressure;
            oldsamplePressure.timestamp = samplesPressure.at(sampleIndexPressure).timestamp;
            oldsamplePressure.pressure = samplesPressure.at(sampleIndexPressure).pressure;
            pressureRate = (samplePressure.pressure - oldsamplePressure.pressure) / (samplePressure.timestamp - oldsamplePressure.timestamp).toSecondsDouble();
        }

        //compute angular accceleration Rate
        float accelerationRate = 0;
        AccelerationRate sampleAcceleration;
        sampleAcceleration.timestamp = now;
        sampleAcceleration.rotationRate = rotationRate;
        if (samplesAcceleration.size() < maxSamplesAcceleration)
        {
            samplesAcceleration.push_back(sampleAcceleration);
        }
        else
        {
            samplesAcceleration[sampleIndexAcceleration] = sampleAcceleration;
            sampleIndexAcceleration = (sampleIndexAcceleration + 1) % maxSamplesAcceleration;
            AccelerationRate oldsampleAcceleration;
            oldsampleAcceleration.timestamp = samplesAcceleration.at(sampleIndexAcceleration).timestamp;
            oldsampleAcceleration.rotationRate = samplesAcceleration.at(sampleIndexAcceleration).rotationRate;
            accelerationRate = (sampleAcceleration.rotationRate - oldsampleAcceleration.rotationRate) / (sampleAcceleration.timestamp - oldsampleAcceleration.timestamp).toSecondsDouble();
        }
        if ((abs(pressureRate) > 50))
        {
            ARMARX_IMPORTANT << "contact";
        }

        if (topicPrx)
        {
            topicPrx->reportSensorValues(data.id, data.pressure, data.qw, data.qx, data.qy, data.qz, pressureRate, rotationRate, accelerationRate, data.accelx, data.accely, data.accelz, nowTimestamp);
        }
    }
}

// get imu values from incoming string
OrientedTactileSensorUnit::SensorData OrientedTactileSensorUnit::getValues(std::string line)
{
    SensorData data;
    std::vector<std::string> splitValues;
    boost::split(splitValues, line, boost::is_any_of(" "));
    data.id = stoi(splitValues.at(0));
    data.pressure = std::stof(splitValues.at(1));
    data.qw = std::stof(splitValues.at(2));
    data.qx = std::stof(splitValues.at(3));
    data.qy = std::stof(splitValues.at(4));
    data.qz = std::stof(splitValues.at(5));
    data.accelx = std::stof(splitValues.at(6));
    data.accely = std::stof(splitValues.at(7));
    data.accelz = std::stof(splitValues.at(8));
    return data;
}

std::string space = " ";
bool OrientedTactileSensorUnit::loadCalibration()
{
    std::string calibrationStream = getProperty<std::string>("CalibrationData").getValue();
    std::string arduinoLine;
    while (arduinoLine.find("mode") == std::string::npos)
    {
        getline(arduino, arduinoLine, '\n');
    }
    arduino.close();

    arduino.open(getProperty<std::string>("SerialInterfaceDevice").getValue(), std::ios::out);
    arduino << "load";
    arduino.flush();
    arduino.close();

    arduino.open(getProperty<std::string>("SerialInterfaceDevice").getValue(), std::ios::in);
    while (arduinoLine.find("calibration data") == std::string::npos)
    {
        getline(arduino, arduinoLine, '\n');
    }
    arduino.close();

    arduino.open(getProperty<std::string>("SerialInterfaceDevice").getValue(), std::ios::out);
    arduino << calibrationStream;
    arduino.flush();
    arduino.close();

    arduino.open(getProperty<std::string>("SerialInterfaceDevice").getValue(), std::ios::in);

    while (arduinoLine.find("Calibration Sucessfull") == std::string::npos)
    {
        getline(arduino, arduinoLine, '\n');
    }
    return true;
}

bool OrientedTactileSensorUnit::getCalibrationValues(std::string line)
{
    std::vector<std::string> splitValues;
    boost::split(splitValues, line, boost::is_any_of(" "));
    calibration.accel_offset_x = stoi(splitValues.at(0));
    calibration.accel_offset_y = stoi(splitValues.at(1));
    calibration.accel_offset_z = stoi(splitValues.at(2));
    calibration.gyro_offset_x = stoi(splitValues.at(3));
    calibration.gyro_offset_y = stoi(splitValues.at(4));
    calibration.gyro_offset_z = stoi(splitValues.at(5));
    calibration.mag_offset_x = stoi(splitValues.at(6));
    calibration.mag_offset_y = stoi(splitValues.at(7));
    calibration.mag_offset_z = stoi(splitValues.at(8));
    calibration.accel_radius = stoi(splitValues.at(9));
    calibration.mag_radius = stoi(splitValues.at(10));
    std::string space = " ";
    std::string calibrationStream = "";
    calibrationStream = calibrationStream + std::to_string(calibration.accel_offset_x) + space + std::to_string(calibration.accel_offset_y) + space + std::to_string(calibration.accel_offset_z) + space + std::to_string(calibration.gyro_offset_x) + space + std::to_string(calibration.gyro_offset_y) + space +
                        std::to_string(calibration.gyro_offset_z) + space + std::to_string(calibration.mag_offset_x) + space + std::to_string(calibration.mag_offset_y) + space + std::to_string(calibration.mag_offset_z) + space + std::to_string(calibration.accel_radius) + space + std::to_string(calibration.mag_radius) + space;
    ARMARX_IMPORTANT << "calibration data: " << calibrationStream ;
    return true;
}


