/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "BasicControllers.h"

#include <boost/algorithm/clamp.hpp>

#include <ArmarXCore/core/util/algorithm.h>

///REMOVE
#include <iostream>

namespace armarx
{
    float velocityControlWithAccelerationBounds(
        float dt, float maxDt,
        const float currentV, float targetV, float maxV,
        float acceleration, float deceleration,
        float directSetVLimit
    )
    {
        dt = std::min(std::abs(dt), std::abs(maxDt));
        maxV = std::abs(maxV);
        acceleration = std::abs(acceleration);
        deceleration = std::abs(deceleration);
        targetV = boost::algorithm::clamp(targetV, -maxV, maxV);

        //we can have 3 cases:
        // 1. we directly set v and ignore acc/dec (if |curr - target| <= limit)
        // 2. we need to accelerate                (if curr and target v have same sign and |curr| < |target|)
        // 3. we need to decellerate               (other cases)

        //handle case 1
        const float curverror=targetV - currentV;
        if (std::abs(curverror) < directSetVLimit)
        {
            return targetV;
        }

        //handle case 2 + 3
        const bool accelerate = sign(targetV) == sign(currentV) &&     // v in same direction
                    std::abs(targetV) > std::abs(currentV); // currently to slow

        const float usedacc = accelerate? acceleration:-deceleration;
        const float maxDeltaV = std::abs(usedacc * dt);
        if(maxDeltaV >= std::abs(curverror))
        {
            //we change our v too much with the used acceleration
            //but we could reach our target with a lower acceleration ->set the target

            return targetV;
        }
        const float deltaVel = boost::algorithm::clamp(sign(currentV) * usedacc * dt, -maxDeltaV, maxDeltaV);
        const float nextV = currentV + deltaVel;
        return nextV;
    }

////////////////////////////
//wip?
    float velocityControlWithAccelerationAndPositionBounds(
        float dt, float maxDt,
        float currentV, float targetV, float maxV,
        float acceleration, float deceleration,
        float directSetVLimit,
        float currentPosition,
        float positionLimitLoSoft, float positionLimitHiSoft,
        float positionSoftLimitViolationVelocity,
        float positionLimitLoHard, float positionLimitHiHard)
    {
        if(currentPosition <= positionLimitLoHard || currentPosition >= positionLimitHiHard)
        {
            return std::nanf("1");
        }

        float softLimitViolation =0;
        if (currentPosition <= positionLimitLoSoft)
        {
            softLimitViolation = -1;
        }
        if (currentPosition >= positionLimitHiSoft)
        {
            softLimitViolation = 1;
        }

        const float upperDt = std::max(std::abs(dt), std::abs(maxDt));
        dt = std::min(std::abs(dt), std::abs(maxDt));
        maxV = std::abs(maxV);

        //we can have 4 cases:
        // 1. we need to decelerate now or we will violate the position limits (maybe we still will violate them, e.g. if we violated them initially)
        // 2. we directly set v and ignore acc/dec (if |curr - target| <= limit)
        // 3. we need to accelerate                (if curr and target v have same sign and |curr| < |target|)
        // 4. we need to decellerate               (other cases)
        float nextv;
        //handle case 1
        const float vsquared = currentV * currentV;
        const float brakingDist = sign(currentV) * vsquared / 2.f / std::abs(deceleration); //the braking distance points in the direction of the velocity


        const float posIfBrakingNow = currentPosition + brakingDist;
        if (posIfBrakingNow <= positionLimitLoSoft || posIfBrakingNow >= positionLimitHiSoft)
        {
            //case 1. -> brake now! (we try to have v=0 at the limit)
            const auto limit = posIfBrakingNow <= positionLimitLoSoft ? positionLimitLoSoft : positionLimitHiSoft;
            const float wayToGo = limit - currentPosition;

            //decellerate!
            // s = v²/(2a)  <=>  a = v²/(2s)
            const float dec = std::abs(vsquared / 2.f / wayToGo);
            const float vel = currentV - sign(currentV) * dec * upperDt;
            nextv = boost::algorithm::clamp(vel,-maxV,maxV);
            if(sign(currentV) != sign(nextv))
            {
                //stop now
                nextv = 0;
            }
        }
        else
        {
            //handle 2-3
            nextv = velocityControlWithAccelerationBounds(dt, maxDt, currentV, targetV, maxV, acceleration, deceleration, directSetVLimit);
        }
        if(softLimitViolation == sign(nextv))
        {
            //the area between soft and hard limits is sticky
            //the controller can only move out of it (not further in)
            return 0;
        }

        //the next velocity will not violate the pos bounds harder than they are already
        return nextv;
    }


    float positionThroughVelocityControlWithAccelerationBounds(
        float dt, float maxDt,
        float currentV, float maxV,
        float acceleration, float deceleration,
        float currentPosition, float targetPosition,
        float pControlPosErrorLimit, float pControlVelLimit, float p
    )
    {
        dt = std::min(std::abs(dt), std::abs(maxDt));
        maxV = std::abs(maxV);
        acceleration = std::abs(acceleration);
        deceleration = std::abs(deceleration);
        pControlPosErrorLimit = std::abs(pControlPosErrorLimit);
        pControlVelLimit = std::abs(pControlVelLimit);
        const float signV = sign(currentV);
        //we can have 3 cases:
        // 1. we use a p controller and ignore acc/dec (if |currentPosition - targetPosition| <= pControlPosErrorLimit AND |currentV| < pControlVelLimit)
        // 2. we need to accelerate (or hold vel)      (if e = (targetPosition - currentPosition)
        //                                                 the brakingDistance have the same sign and brakingDistance < e
        //                                                 and currentVel <= maxV)
        // 3. we need to decellerate                   (other cases)

        //handle case 1
        const float positionError = targetPosition - currentPosition;
        if (std::abs(positionError) < pControlPosErrorLimit && std::abs(currentV) < pControlVelLimit)
        {
            return positionError * p;
        }

        //handle case 2-3
        const float brakingDistance = signV * currentV * currentV / 2.f / deceleration; //the braking distance points in the direction of the velocity
        const float posIfBrakingNow = currentPosition + brakingDistance;
        const float posErrorIfBrakingNow = targetPosition - posIfBrakingNow;

        const bool decelerate =
                std::abs(currentV) > maxV || // we need to slow down (to stay in [-maxV,maxV]
                std::abs(posIfBrakingNow-targetPosition) <= pControlPosErrorLimit || // we want to hit the target
                sign(posErrorIfBrakingNow) != signV;  // we are moving away from the target
        const float usedacc= decelerate ?  -deceleration : acceleration;
        const float maxDeltaV = std::abs(usedacc * dt);
        const float deltaVel = boost::algorithm::clamp(signV * usedacc * dt, -maxDeltaV, maxDeltaV);
        return boost::algorithm::clamp(currentV + deltaVel, -maxV, maxV);
    }

    float positionThroughVelocityControlWithAccelerationAndPositionBounds(
        float dt, float maxDt,
        float currentV, float maxV,
        float acceleration, float deceleration,
        float currentPosition, float targetPosition,
        float pControlPosErrorLimit, float pControlVelLimit, float p,
        float positionLimitLo, float positionLimitHi

    )
    {
        dt = std::min(std::abs(dt), std::abs(maxDt));
        maxV = std::abs(maxV);
        acceleration = std::abs(acceleration);
        deceleration = std::abs(deceleration);
        pControlPosErrorLimit = std::abs(pControlPosErrorLimit);
        const float signV = sign(currentV);
        //we can have 4 cases:
        // 1. we need to decelerate now or we will violate the position limits (maybe we still will violate them)
        // 2. we use a p controller and ignore acc/dec (if |currentPosition - targetPosition| <= pControlPosErrorLimit AND |currentPosition - targetPosition|*p < maxV)
        // 4. we need to accelerate (or hold vel)      (if e = (targetPosition - currentPosition)
        //                                                 and the brakingDistance have the same sign and brakingDistance < e
        //                                                 and currentVel <= maxV)
        // 5. we need to decellerate                   (other cases)

        //handle case 1
        const float vsquared = currentV * currentV;
        const float brakingDistance = signV * vsquared / 2.f / deceleration; //the braking distance points in the direction of the velocity
        const float posIfBrakingNow = currentPosition + brakingDistance;
        if (posIfBrakingNow <= positionLimitLo || posIfBrakingNow >= positionLimitHi)
        {
            //case 1. -> brake now! (we try to have v=0 at the limit)
            const auto limit = brakingDistance > 0 ? positionLimitHi : positionLimitLo;
            const float wayToGo = std::abs(limit - currentPosition);
            // s = v²/(2a)  <=>  a = v²/(2s)
            const float dec = std::abs(vsquared / 2.f / wayToGo);
            const float vel = currentV - signV * dec * dt;
            return vel;
        }

        //handle case 2-3
        return positionThroughVelocityControlWithAccelerationBounds(
            dt, maxDt,
            currentV, maxV,
            acceleration, deceleration,
            currentPosition, boost::algorithm::clamp(targetPosition, positionLimitLo, positionLimitHi),
            pControlPosErrorLimit, pControlVelLimit, p
        );
    }

    ////////////////////////////
    //wip


    float positionThroughVelocityControlWithAccelerationBoundsAndPeriodicPosition(
            float dt, float maxDt,
            float currentV, float maxV,
            float acceleration, float deceleration,
            float currentPosition, float targetPosition,
            float pControlPosErrorLimit, float pControlVelLimit, float p,
            float &direction,
            float positionPeriodLo, float positionPeriodHi)
    {
        currentPosition = periodicClamp(currentPosition, positionPeriodLo, positionPeriodHi);
        targetPosition  = periodicClamp(targetPosition , positionPeriodLo, positionPeriodHi);

        const float brakingDist = brakingDistance(currentV, deceleration);
        const float posIfBrakingNow = currentPosition + brakingDist;
        const float posIfBrakingNowError = targetPosition - posIfBrakingNow;
        const float posError = targetPosition - currentPosition;
        if(
                std::abs(posIfBrakingNowError) <= pControlPosErrorLimit ||
                std::abs(posError) <= pControlPosErrorLimit
        )
        {
            //this allows slight overshooting (in the limits of the p controller)
            return positionThroughVelocityControlWithAccelerationBounds(
                        dt, maxDt,
                        currentV, maxV,
                        acceleration, deceleration,
                        currentPosition, targetPosition,
                        pControlPosErrorLimit, pControlVelLimit, p
                    );
        }

        //we transform the problem with periodic bounds into
        //a problem with position bounds
        const float positionPeriodLength  = std::abs(positionPeriodHi - positionPeriodLo);

        //we shift the positions such that 0 == target
        currentPosition = periodicClamp(currentPosition - targetPosition,  0.f, positionPeriodLength);
        //how many times will we go ovet the target if we simply bake now?
        const float overshoot = std::trunc((currentPosition + brakingDist)/ positionPeriodLength);

        if(true||direction == 0)
        {
            //determine the direction to go (1 == pos vel, -1 == neg vel)
            direction = (periodicClamp(currentPosition + brakingDist, 0.f , positionPeriodLength) >= positionPeriodLength/2) ? 1: -1;
        }
        //shift the target away from 0
        targetPosition = (overshoot-std::min(0.f,-direction)) * positionPeriodLength;// - direction * pControlPosErrorLimit;

        //move
        return positionThroughVelocityControlWithAccelerationBounds(
            dt, maxDt,
            currentV, maxV,
            acceleration, deceleration,
            currentPosition, targetPosition,
            pControlPosErrorLimit, pControlVelLimit, p
        );
    }


}
