/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_LIB_RobotAPI_BasicControllers_H
#define _ARMARX_LIB_RobotAPI_BasicControllers_H

#include <cmath>
#include <ArmarXCore/core/util/algorithm.h>

namespace armarx
{
    /**
     * @param v0 The initial velocity.
     * @param deceleration The deceleration.
     * @return The braking distance given the parameters.
     */
    inline float brakingDistance(float v0, float deceleration)
    {
        return sign(v0)*std::pow(v0, 2.f) / 2.f / std::abs(deceleration);
    }

    template<class T>
    T periodicClamp(T value, T periodLo, T periodHi)
    {
        float dist = periodHi - periodLo;
        return std::fmod(value-periodLo + dist, dist) + periodLo;
    }

    float velocityControlWithAccelerationBounds(
        float dt, float maxDt,
        const float currentV, float targetV, float maxV,
        float acceleration, float deceleration,
        float directSetVLimit);
    /**
     * @brief Performs velocity control while staying in positional bounds, obeying acceleration / deceleration and staying below a velocity bound.
     *
     * we can have 4 cases:
     * 1. we need to decelerate now or we will violate the position limits (maybe we still will violate them. e.f. if we already violate them or the robot can't brake fast enough)
     * 2. we directly set v and ignore acc/dec (if |currentV - targetV| <= directSetVLimit)
     * 3. we need to accelerate                (if currentV and targetV have same sign and |currentV| < |currentV|)
     * 4. we need to decellerate               (other cases)
     * @param dt The time in seconds until the next call is made. (use the time since the last call as an approximate)
     * @param maxDt Limits dt in case the given approximation has a sudden high value.
     * @param currentV
     * @param targetV
     * @param maxV
     * @param acceleration
     * @param deceleration
     * @param directSetVLimit In case |currentV - targetV| <= directSetVLimit this function will return targetV (if the position bounds are not violated)
     * @param currentPosition
     * @param positionLimitLo
     * @param positionLimitHi
     * @return The next velocity.
     */
    float velocityControlWithAccelerationAndPositionBounds(float dt, float maxDt,
        float currentV, float targetV, float maxV,
        float acceleration, float deceleration,
        float directSetVLimit,
        float currentPosition,
        float positionLimitLoSoft, float positionLimitHiSoft,
        float positionSoftLimitViolationVelocity,
        float positionLimitLoHard, float positionLimitHiHard);

    float positionThroughVelocityControlWithAccelerationBounds(float dt, float maxDt,
        float currentV, float maxV,
        float acceleration, float deceleration,
        float currentPosition, float targetPosition,
        float pControlPosErrorLimit, float pControlVelLimit, float p
    );

    float positionThroughVelocityControlWithAccelerationAndPositionBounds(
        float dt, float maxDt,
        float currentV, float maxV,
        float acceleration, float deceleration,
        float currentPosition, float targetPosition,
            float pControlPosErrorLimit, float pControlVelLimit, float p,
        float positionLimitLo, float positionLimitHi);

    //    float positionThroughVelocityControlWithAccelerationBounds(
    //            float dt, float maxDt,
    //            float currentV, float maxV,
    //            float acceleration, float deceleration,
    //            float currentPosition, float targetPosition,
    //            float pControlPosErrorLimit, float p);



    float positionThroughVelocityControlWithAccelerationBoundsAndPeriodicPosition(float dt, float maxDt,
            float currentV, float maxV,
            float acceleration, float deceleration,
            float currentPosition, float targetPosition,
            float pControlPosErrorLimit, float pControlVelLimit, float p,
            float& direction,
            float positionPeriodLo, float positionPeriodHi);
}

#endif
