/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::LVL1KinematicUnitPositionController
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "LVL1KinematicUnitPositionController.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../DataUnits/KinematicDataUnit.h"
#include "../BasicControllers.h"
#include <RobotAPI/interface/libraries/RTRobotUnit/LVL1Controllers/LVL1KinematicUnitControllerConfigs.h>
using namespace armarx;

LVL1KinematicUnitPositionController::LVL1KinematicUnitPositionController(LVL1ControllerDataProviderInterfacePtr prov, LVL1ControllerConfigPtr config)
{
    LVL1KinematicUnitPositionControllerConfigPtr cfg = LVL1KinematicUnitPositionControllerConfigPtr::dynamicCast(config);
    ARMARX_CHECK_EXPRESSION_W_HINT(cfg,
                                   "The provided config has the wrong type! The type is " << config->ice_id()
                                   << " instead of " << LVL1KinematicUnitPositionControllerConfig::ice_staticId());
    //make sure the used units are supported
    auto kinunit = prov->getRTKinematicDataUnit();
    ARMARX_CHECK_EXPRESSION_W_HINT(kinunit, "The DataProvider " << prov->getName() << " has no kinematic data unit");
    //get pointers to sensor values from units
    std::vector<const float*>&& jointPos = kinunit->getJointAngles({cfg->jointName});
    ARMARX_CHECK_EXPRESSION(jointPos.size() == 1);
    ARMARX_CHECK_EXPRESSION(jointPos.at(0));
    currentPosition = jointPos.at(0);
    std::vector<const float*> jointVel = kinunit->getJointVelocities({cfg->jointName});
    ARMARX_CHECK_EXPRESSION(jointVel.size() == 1);
    ARMARX_CHECK_EXPRESSION(jointVel.at(0));
    currentVelocity = jointVel.at(0);
    velocityTarget = dynamic_cast<ActuatorVelocityTarget*>(prov->getJointTarget(cfg->jointName, ControlModes::VelocityMode));
    ARMARX_CHECK_EXPRESSION_W_HINT(velocityTarget, "The joint " << cfg->jointName << " has no controll mode " << ControlModes::VelocityMode);
    //read rest of config
    initialSettings.maxVelocity         = std::abs(cfg->maxVelocity);
    initialSettings.acceleration        = std::abs(cfg->acceleration);
    initialSettings.deceleration        = std::abs(cfg->deceleration);
    initialSettings.directPControlLimit = std::abs(cfg->directPControlLimit);
    initialSettings.p                   = std::abs(cfg->p);
    initialSettings.maxDt               = std::abs(cfg->maxDt);
    initialSettings.limitPositionLo     = cfg->limitPositionLo;
    initialSettings.limitPositionHi     = cfg->limitPositionHi;
    ARMARX_CHECK_EXPRESSION(initialSettings.limitPositionLo <= initialSettings.limitPositionHi);
    initialSettings.positionTargetValue = cfg->initialPositionTargetValue;
    reinitTripleBuffer(initialSettings);
}

void LVL1KinematicUnitPositionController::rtRun(const IceUtil::Time&, const IceUtil::Time& timeSinceLastIteration)
{
    //    auto& ct = rtGetControlStruct();
    //    velocityTarget->velocity = positionThroughVelocityControlWithAccelerationAndPositionBounds(
    //                                   /*dt*/ timeSinceLastIteration.toSecondsDouble(), /*maxDt*/ ct.maxDt,
    //                                   *currentVelocity, ct.maxVelocity,
    //                                   ct.acceleration, ct.deceleration,
    //                                   *currentPosition, ct.positionTargetValue,
    //                                   ct.directPControlLimit, ct.p, ct.limitPositionLo, ct.limitPositionHi
    //                               );
}

void LVL1KinematicUnitPositionController::setTargetPosition(float pos, float maxVelocity, float acceleration, float deceleration)
{
    LockGuardType l {controlDataMutex};
    getWriterControlStruct().positionTargetValue = pos;
    getWriterControlStruct().maxVelocity  = std::abs(maxVelocity);
    getWriterControlStruct().acceleration = std::abs(acceleration);
    getWriterControlStruct().deceleration = std::abs(deceleration);
    writeControlStruct();
}

void LVL1KinematicUnitPositionController::setTargetPosition(float pos)
{
    LockGuardType l {controlDataMutex};
    getWriterControlStruct().positionTargetValue = pos;
    writeControlStruct();
}

void LVL1KinematicUnitPositionController::setControlSettings(LVL1KinematicUnitPositionControllerControlSettings settings)
{
    settings.maxVelocity          = std::abs(settings.maxVelocity);
    settings.acceleration         = std::abs(settings.acceleration);
    settings.deceleration         = std::abs(settings.deceleration);
    settings.directPControlLimit  = std::abs(settings.directPControlLimit);
    settings.p                    = std::abs(settings.p);
    settings.maxDt           = std::abs(settings.maxDt);
    ARMARX_CHECK_EXPRESSION(settings.limitPositionLo <= settings.limitPositionHi);
    setControlStruct(settings);
}

LVL1ControllerRegistration<LVL1KinematicUnitPositionController> registrationLVL1KinematicUnitPositionController("LVL1KinematicUnitPositionController");
