/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::RobotUnit
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "RobotUnit.h"

#include <ArmarXCore/core/system/cmake/CMakePackageFinder.h>

#include <ArmarXCore/core/system/DynamicLibrary.h>

armarx::PropertyDefinitionsPtr armarx::RobotUnit::createPropertyDefinitions()
{
    return armarx::PropertyDefinitionsPtr(new RobotUnitPropertyDefinitions(getConfigIdentifier()));
}

void armarx::RobotUnit::addLVL0Controller(const std::string& jointName, armarx::LVL0Controller& lvl0Controller)
{
    std::string controlMode = lvl0Controller.getControlMode();
    GuardType guard {dataMutex};
    ARMARX_CHECK_EXPRESSION_W_HINT(areJointNamesSet(), "Joint names a were not set!");
    if (hasLVL0Controller(jointName, controlMode))
    {
        ARMARX_ERROR << "A LVL0Controller for " + jointName + "(" + controlMode + ") does already exist!";
        throw std::invalid_argument {"A LVL0Controller for " + jointName + "(" + controlMode + ") does already exist!"};
    }
    lvl0Controllers[jointName][controlMode] = &lvl0Controller;
    if (controlMode == ControlModes::EmergencyStopMode)
    {
        lvl0ControllersEmergencyStop.at(jointNameIndices.at(jointName)) = &lvl0Controller;
    }
}

const armarx::LVL0Controller* armarx::RobotUnit::getLVL0Controller(const std::string& jointName, const std::string& controlMode) const
{
    GuardType guard {dataMutex};
    ARMARX_CHECK_EXPRESSION_W_HINT(areJointNamesSet(), "Joint names a were not set!");
    return lvl0Controllers.at(jointName).at(controlMode);
}

armarx::LVL0Controller* armarx::RobotUnit::getLVL0Controller(const std::string& jointName, const std::string& controlMode)
{
    GuardType guard {dataMutex};
    ARMARX_CHECK_EXPRESSION_W_HINT(areJointNamesSet(), "Joint names a were not set!");
    return lvl0Controllers.at(jointName).at(controlMode);
}

Ice::StringSeq armarx::RobotUnit::getControllersKnown(const Ice::Current&) const
{
    return LVL1ControllerRegistry::getKeys();
}

Ice::StringSeq armarx::RobotUnit::getControllerNamesLoaded(const Ice::Current&) const
{
    GuardType guard {dataMutex};
    Ice::StringSeq result;
    result.reserve(lvl1Controllers.size());
    for (const auto& lvl1 : lvl1Controllers)
    {
        result.emplace_back(lvl1.first);
    }
    return result;
}
Ice::StringSeq armarx::RobotUnit::getControllerNamesRequested(const Ice::Current&) const
{
    GuardType guard {dataMutex};
    Ice::StringSeq result;
    result.reserve(lvl1Controllers.size());
    for (const auto& lvl1 : getRequestedLVL1Controllers())
    {
        if (lvl1)
        {
            result.emplace_back(lvl1->getName());
        }
    }
    return result;
}
Ice::StringSeq armarx::RobotUnit::getControllerNamesActivated(const Ice::Current&) const
{
    GuardType guard {dataMutex};
    Ice::StringSeq result;
    result.reserve(lvl1Controllers.size());
    for (const auto& lvl1 : getActivatedLVL1Controllers())
    {
        if (!lvl1)
        {
            continue;
        }
        result.emplace_back(lvl1->getName());
    }
    return result;
}

armarx::StringLVL1ControllerPrxDictionary armarx::RobotUnit::getControllersLoaded(const Ice::Current&) const
{
    GuardType guard {dataMutex};
    StringLVL1ControllerPrxDictionary result;
    for (const auto& lvl1 : lvl1Controllers)
    {
        result[lvl1.first] = LVL1ControllerInterfacePrx::uncheckedCast(lvl1.second->getProxy());
    }
    return result;
}
armarx::StringLVL1ControllerPrxDictionary armarx::RobotUnit::getControllersRequested(const Ice::Current&) const
{
    GuardType guard {dataMutex};
    StringLVL1ControllerPrxDictionary result;
    for (const auto& lvl1 : getRequestedLVL1Controllers())
    {
        if (lvl1)
        {
            result[lvl1->getName()] = LVL1ControllerInterfacePrx::uncheckedCast(lvl1->getProxy());
        }
    }
    return result;
}
armarx::StringLVL1ControllerPrxDictionary armarx::RobotUnit::getControllersActivated(const Ice::Current&) const
{
    GuardType guard {dataMutex};
    StringLVL1ControllerPrxDictionary result;
    for (const auto& lvl1 : getActivatedLVL1Controllers())
    {
        result[lvl1->getName()] = LVL1ControllerInterfacePrx::uncheckedCast(lvl1->getProxy());
    }
    return result;
}

armarx::JointNameToLVL1Dictionary armarx::RobotUnit::getControllerJointAssignment(const Ice::Current&) const
{
    GuardType guard {dataMutex};
    JointNameToLVL1Dictionary result;
    for (const auto& joint : jointNames)
    {
        result[joint] = "";
    }
    for (const auto& lvl1 : getActivatedLVL1Controllers())
    {
        if (!lvl1)
        {
            continue;
        }
        for (const auto& jointMode : lvl1->jointControlModeMap)
        {
            result[jointMode.first] = lvl1->getName();
        }
    }
    return result;
}

armarx::JointNameToControlModeDictionary armarx::RobotUnit::getJointControlModesRequested(const Ice::Current&) const
{
    GuardType guard {dataMutex};
    JointNameToControlModeDictionary result;
    const auto& requestedModes = getRequestedLVL0Controllers();
    ARMARX_CHECK_AND_THROW(jointNames.size() == requestedModes.size(), std::logic_error);
    for (std::size_t i = 0; i < jointNames.size(); ++i)
    {
        if (requestedModes.at(i))
        {
            result[jointNames.at(i)] = requestedModes.at(i)->getControlMode();
        }
        else
        {
            result[jointNames.at(i)] = "<NO LVL0CONTROLLER SET>";
        }
    }
    return result;
}
armarx::JointNameToControlModeDictionary armarx::RobotUnit::getJointControlModesActivated(const Ice::Current&) const
{
    GuardType guard {dataMutex};
    JointNameToControlModeDictionary result;
    const auto& activatedModes = getActivatedLVL0Controllers();
    ARMARX_CHECK_AND_THROW(jointNames.size() == activatedModes.size(), std::logic_error);
    for (std::size_t i = 0; i < jointNames.size(); ++i)
    {
        if (activatedModes.at(i))
        {
            result[jointNames.at(i)] = activatedModes.at(i)->getControlMode();
        }
        else
        {
            result[jointNames.at(i)] = "<NO LVL0CONTROLLER SET>";
        }

    }
    return result;
}
armarx::JointNameToControlModesDictionary armarx::RobotUnit::getJointControlModesSupported(const Ice::Current&) const
{
    GuardType guard {dataMutex};
    JointNameToControlModesDictionary result;
    for (const auto& joint : lvl0Controllers)
    {
        std::vector<std::string> modes;
        modes.reserve(joint.second.size());
        for (const auto& mode : joint.second)
        {
            modes.emplace_back(mode.first);
        }
        result[joint.first] = std::move(modes);
    }
    return result;
}

void armarx::RobotUnit::switchSetup(const Ice::StringSeq& controllerRequestedNames, const Ice::Current&)
{
    GuardType guard {dataMutex};
    ARMARX_CHECK_EXPRESSION_W_HINT(areJointNamesSet(), "Joint names a were not set!");
    //check if all of these controllers exist
    for (const auto& lvl1 : controllerRequestedNames)
    {
        if (!hasLVL1Controller(lvl1))
        {
            throw InvalidArgumentException {"No controller of the name '" + lvl1 + "' is loaded!"};
        }
    }
    //check controllers (is there a collision/which controllers need to be deactivated)
    const auto currentActiveLVL1Controllers = getControllerNamesActivated();
    auto lvl1ControllerAssignement = getControllerJointAssignment();

    ARMARX_INFO << "current active lvl1 controllers:\n" << currentActiveLVL1Controllers;

    std::set<std::string> controllersToActivate {currentActiveLVL1Controllers.begin(), currentActiveLVL1Controllers.end()}; //these controllers will be set as active controllers

    for (const auto& toActivate : controllerRequestedNames)
    {
        if (controllersToActivate.count(toActivate))
        {
            ARMARX_INFO << "controller already active: " << toActivate;
            continue;
        }
        controllersToActivate.emplace(toActivate);

        ARMARX_INFO << "activate '" << toActivate << "'";

        const auto& lvl1 = lvl1Controllers.at(toActivate);
        //check and update the assignement map
        for (const auto& jointControlMode : lvl1->jointControlModeMap)
        {
            const auto& joint = jointControlMode.first;
            const auto& currentAssignedLVL1 = lvl1ControllerAssignement.at(joint);
            if (toActivate == currentAssignedLVL1)
            {
                continue;
            }
            if (currentAssignedLVL1.empty())
            {
                lvl1ControllerAssignement[joint] = toActivate;
                continue;
            }
            //deactivate other controller
            for (auto& assignement : lvl1ControllerAssignement)
            {
                if (assignement.second == currentAssignedLVL1)
                {
                    assignement.second.clear();
                }
            }
            controllersToActivate.erase(currentAssignedLVL1);
            ARMARX_INFO << "deactivated '" << currentAssignedLVL1 << "' (caused by activation of '" << toActivate << "')";
            lvl1ControllerAssignement[joint] = toActivate;
        }
    }
    //verify (exception for collision of requested lvl1)
    for (const auto& toActivate : controllerRequestedNames)
    {
        if (!controllersToActivate.count(toActivate))
        {
            switchSetupError = true;
            ARMARX_ERROR << "The given set of controllers is in conflict and could not be activated! (conflicting is: " + toActivate + ")";
            throw InvalidArgumentException {"The given set of controllers is in conflict and could not be activated! (conflicting is: " + toActivate + ")"};
        }
    }
    //verify (warn for orphant joints) + populate controllersRequested.lvl0Controllers
    getRequestedLVL0Controllers().resize(jointNames.size());
    for (std::size_t i = 0; i < jointNames.size(); ++i)
    {
        const auto& joint = jointNames.at(i);
        if (lvl1ControllerAssignement[joint].empty())
        {
            getRequestedLVL0Controllers().at(i) = getLVL0Controller(joint, ControlModes::EmergencyStopMode);
            ARMARX_WARNING << "Joint '" << joint << "' has no lvl1 controller assigned!";
            continue;
        }
        const auto& lvl0Mode = lvl1Controllers.at(lvl1ControllerAssignement[joint])->jointControlModeMap.at(joint);
        getRequestedLVL0Controllers().at(i) = getLVL0Controller(joint, lvl0Mode);
    }
    //populate controllersRequested.lvl1Controllers
    getRequestedLVL1Controllers().clear();
    getRequestedLVL1Controllers().reserve(controllersToActivate.size());
    for (const auto& lvl1 : controllersToActivate)
    {
        getRequestedLVL1Controllers().emplace_back(lvl1Controllers.at(lvl1));
    }

    //now change the assignement
    controllersRequested.commitWrite();
    ARMARX_INFO << "requested lvl1 controllers:\n" << getControllerNamesRequested();
}

armarx::LVL1ControllerInterfacePrx armarx::RobotUnit::loadController(const std::string& className, const std::string& instanceName, const armarx::LVL1ControllerConfigPtr& config, const Ice::Current&)
{
    if (instanceName.empty())
    {
        ARMARX_ERROR << "The instance name is empty!";
        throw InvalidArgumentException {"The instance name is empty!"};
    }
    //check if we would be able to create the class
    if (!LVL1ControllerRegistry::has(className))
    {
        std::stringstream ss;
        ss << "Requested controller class '" << className << "' unknown! Known classes:" << LVL1ControllerRegistry::getKeys();
        ARMARX_ERROR << ss.str();
        throw InvalidArgumentException {ss.str()};
    }
    auto factory = LVL1ControllerRegistry::get(className);

    GuardType guard {dataMutex};
    //check if the instance name is already in use
    if (hasLVL1Controller(instanceName))
    {
        std::stringstream ss;
        ss << "There already is a controller instance with the name '" << instanceName << "'";
        ARMARX_ERROR << ss.str();
        throw InvalidArgumentException {ss.str()};
    }
    //create the controller
    jointsUsedByLVL1Controller.clear();
    RobotUnitInterfacePtr::dynamicCast(RobotUnitPtr {this});
    LVL1ControllerDataProviderInterfacePtr::dynamicCast(RobotUnitPtr {this});
    LVL1ControllerPtr lvl1 = factory(LVL1ControllerDataProviderInterfacePtr::dynamicCast(RobotUnitPtr {this}), config);
    lvl1->jointControlModeMap = jointsUsedByLVL1Controller;
    lvl1->jointIndices.clear();
    lvl1->jointIndices.reserve(jointsUsedByLVL1Controller.size());
    for (const auto& j : jointsUsedByLVL1Controller)
    {
        lvl1->jointIndices.emplace_back(jointNameIndices.at(j.first));
    }

    getArmarXManager()->addObject(lvl1, instanceName);
    const auto prx = lvl1->getProxy(-1);
    lvl1Controllers[instanceName] = lvl1;
    return LVL1ControllerInterfacePrx::uncheckedCast(prx);
}

armarx::TargetBase* armarx::RobotUnit::getJointTarget(const std::string& jointName, const std::string& controlMode)
{
    GuardType guard {dataMutex};
    if (!hasLVL0Controller(jointName, controlMode))
    {
        return nullptr;
    }
    const auto lvl0 = getLVL0Controller(jointName, controlMode);
    ARMARX_CHECK_EXPRESSION(lvl0->getControlMode() == controlMode);
    jointsUsedByLVL1Controller[jointName] = controlMode;
    return lvl0->getTarget();
}

void armarx::RobotUnit::setJointNames(const std::vector<std::string>& names)
{
    GuardType guard {dataMutex};
    ARMARX_CHECK_EXPRESSION_W_HINT(!areJointNamesSet(), "JointNames were already set!");
    jointNames = names;
    jointNameIndices = toIndexMap(jointNames);
    lvl0ControllersEmergencyStop.resize(jointNames.size());
    controllersRequested.reinitAllBuffers(LVL0AndLVL1ControllerList(jointNames.size()));
    controllersActivated.reinitAllBuffers(LVL0AndLVL1ControllerList(jointNames.size()));
}

bool armarx::RobotUnit::hasLVL1Controller(const std::string& name) const
{
    GuardType guard {dataMutex};
    ARMARX_CHECK_EXPRESSION_W_HINT(areJointNamesSet(), "Joint names a were not set!");
    return lvl1Controllers.end() != lvl1Controllers.find(name);
}

void armarx::RobotUnit::rtSetLVL1ControllerActive(const armarx::LVL1ControllerPtr& lvl1, bool isActive)
{
    if (isActive)
    {
        lvl1->rtActivateController();
    }
    else
    {
        lvl1->rtDeactivateController();
    }
}

bool armarx::RobotUnit::areJointNamesSet() const
{
    GuardType guard {dataMutex};
    return !jointNames.empty();
}

bool armarx::RobotUnit::hasLVL0Controller(const std::string& jointName, const std::string& controlMode) const
{
    GuardType guard {dataMutex};
    ARMARX_CHECK_EXPRESSION_W_HINT(areJointNamesSet(), "Joint names a were not set!");
    return  lvl0Controllers.end()               != lvl0Controllers.find(jointName) &&
            lvl0Controllers.at(jointName).end() != lvl0Controllers.at(jointName).find(controlMode) &&
            lvl0Controllers.at(jointName).at(controlMode);
}


bool armarx::RobotUnit::loadLibFromAbsolutePath(const std::string& path)
{
    GuardType guard {dataMutex};
    if (loadedLibs.count(path) > 0)
    {
        return true;
    }
    DynamicLibraryPtr lib(new DynamicLibrary());
    try
    {
        lib->load(path);
    }
    catch (...)
    {
        handleExceptions();
        return false;
    }

    if (lib->isLibraryLoaded())
    {
        loadedLibs[path] = lib;
    }
    else
    {
        ARMARX_ERROR << "Could not load lib " + path + ": " + lib->getErrorMessage();
        return false;
    }


    return true;
}

bool armarx::RobotUnit::loadLibFromPath(const std::string& path, const Ice::Current&)
{
    std::string absPath;
    if (ArmarXDataPath::getAbsolutePath(path, absPath))
    {
        return loadLibFromAbsolutePath(absPath);
    }
    else
    {
        ARMARX_ERROR << "Could not find library " + path;
        return false;
    }
}

bool armarx::RobotUnit::loadLibFromPackage(const std::string& package, const std::string& name, const Ice::Current&)
{
    CMakePackageFinder finder(package);
    if (!finder.packageFound())
    {
        ARMARX_ERROR << "Could not find package '" << package << "'";
        return false;
    }

    for (auto libDirPath : armarx::Split(finder.getLibraryPaths(), ";"))
    {
        boost::filesystem::path fullPath = libDirPath;
        fullPath /= "lib" + name  + "." + DynamicLibrary::GetSharedLibraryFileExtension();
        if (!boost::filesystem::exists(fullPath))
        {
            fullPath = libDirPath;
            fullPath /= name;
            if (!boost::filesystem::exists(fullPath))
            {
                continue;
            }
        }
        if (loadLibFromAbsolutePath(fullPath.string()))
        {
            return true;
        }
    }
    ARMARX_ERROR << "Could not find library " <<  name << " in package " << package;
    return false;
}

bool armarx::RobotUnit::rtSwitchSetup()
{
    if (!rtControllersWereSwitched())
    {
        return false;
    }

    //handle lvl1
    for (std::size_t i = 0; i < rtGetRequestedLVL1Controllers().size(); ++i)
    {
        //deactivate old
        if (rtGetActivatedLVL1Controller(i))
        {
            rtGetActivatedLVL1Controller(i)->rtDeactivateController();
        }

        //activate new
        if (rtGetRequestedLVL1Controller(i))
        {
            rtGetRequestedLVL1Controller(i)->rtActivateController();
        }
        //update activated
        rtGetActivatedLVL1Controller(i) = rtGetRequestedLVL1Controller(i);
    }
    //handle lvl0
    for (std::size_t i = 0; i < rtGetRequestedLVL0Controllers().size(); ++i)
    {
        if (rtGetRequestedLVL0Controller(i) != rtGetActivatedLVL0Controller(i))
        {
            rtSwitchLVL0Controller(i, rtGetActivatedLVL0Controller(i), rtGetRequestedLVL0Controller(i));
            rtGetActivatedLVL0Controller(i) = rtGetRequestedLVL0Controller(i);
        }

        if (!rtGetRequestedLVL0Controller(i))
        {
            //no lvl0 controller is set!
            rtSwitchLVL0Controller(i, rtGetActivatedLVL0Controller(i), rtGetEmergencyStopLVL0Controller(i));
            rtGetActivatedLVL0Controller(i) = rtGetActivatedLVL0Controller(i);
        }
    }


    return true;
}

void armarx::RobotUnit::rtRunLVL1Controllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
{
    for (LVL1ControllerPtr& lvl1 : rtGetActivatedLVL1Controllers())
    {
        if (!lvl1)
        {
            continue;
        }
        lvl1->rtSwapBufferAndRun(sensorValuesTimestamp, timeSinceLastIteration);
    }
}

void armarx::RobotUnit::rtDeactivateAssignedLVL1Controller(std::size_t index)
{
    for (std::size_t i = 0; i < rtGetRequestedLVL1Controllers().size(); ++i)
    {
        if (!rtGetRequestedLVL1Controller(i))
        {
            continue;
        }
        if (rtGetRequestedLVL1Controller(i)->rtUsesJoint(index))
        {
            rtGetRequestedLVL1Controller(i)->rtDeactivateController();
            for (auto used : rtGetRequestedLVL1Controller(i)->rtGetJointIndices())
            {
                rtSwitchLVL0Controller(used, rtGetActivatedLVL0Controller(i), rtGetEmergencyStopLVL0Controller(i));
                rtGetActivatedLVL0Controller(i) = rtGetEmergencyStopLVL0Controller(i);
            }
            rtGetActivatedLVL1Controller(i) = nullptr;
            return;
        }
    }
}

void armarx::RobotUnit::rtRunLVL0Controllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
{
    for (LVL0Controller* lvl0 : rtGetActivatedLVL0Controllers())
    {
        lvl0->rtRun(sensorValuesTimestamp, timeSinceLastIteration);
    }
}

bool armarx::RobotUnit::validateAddedLVL0Controllers() const
{
    GuardType guard {dataMutex};
    ARMARX_CHECK_EXPRESSION_W_HINT(areJointNamesSet(), "Joint names a were not set!");
    for (const auto& lvl0 : lvl0ControllersEmergencyStop)
    {
        if (!lvl0)
        {
            return false;
        }
    }
    return true;

}
