/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::RobotUnit
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_UNIT_RobotAPI_RobotUnit_H
#define _ARMARX_UNIT_RobotAPI_RobotUnit_H

#include <thread>
#include <atomic>
#include <chrono>
#include <mutex>
#include <unordered_map>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/ArmarXManager.h>
#include <ArmarXCore/core/time/TimeUtil.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/util/algorithm.h>
#include <ArmarXCore/core/util/TripleBuffer.h>

#include <RobotAPI/interface/libraries/RTRobotUnit/RobotUnitInterface.h>
#include "LVL0Controllers/LVL0Controller.h"
#include "LVL1Controllers/LVL1Controller.h"

namespace armarx
{
    class DynamicLibrary;
    typedef boost::shared_ptr<DynamicLibrary> DynamicLibraryPtr;
    /**
     * @class RobotUnitPropertyDefinitions
     * @brief
     */
    class RobotUnitPropertyDefinitions:
        public armarx::ComponentPropertyDefinitions
    {
    public:
        RobotUnitPropertyDefinitions(std::string prefix):
            armarx::ComponentPropertyDefinitions(prefix)
        {
            //defineRequiredProperty<std::string>("PropertyName", "Description");
            //defineOptionalProperty<std::string>("PropertyName", "DefaultValue", "Description");
        }
    };

    /**
     * @defgroup Component-RobotUnit RobotUnit
     * @ingroup RobotAPI-Components
     * A description of the component RobotUnit.
     *
     * @class RobotUnit
     * @ingroup Component-RobotUnit
     * @brief Brief description of class RobotUnit.
     *
     * \section Datastructures you should have
     * \li A thread for RT
     * \li A thread for publishing + errorreporting
     * \li A Tripplebuffer to publish the current robot state + current targets
     * \li A WriteBufferedTripleBuffer containing an error and logging structure (to communicate such information to the publisher thread)
     *
     * \section RobotUnit-tasks-to-implement Tasks to implement when deriving from this class
     * When implementing this class you have to handle the following tasks:
     * \li Implement the real time thread.
     * \li Implement the publisher thread.
     * \li Implement the communication between both threads.
     * \li Load all LVL0 controllers and add them to the unit.
     *
     * \subsection RobotUnit-tasks-to-implement-rt Implementing the RT thread
     * The thread has to be RT.
     * So you must not do anything of the following:
     * \li Call blocking actions (e.g. mutex::lock)
     * \li allocate ram using new (this calls lock on a mutex)
     * \li print output (this may allocate ram or block)
     * \li change the size of a datasructure (strings, vectors, etc) (this may allocate ram)
     *
     * The structure for a thread is:
     * \code{.cpp}
     * void rtThread()
     * {
     *      try
     *      {
     *          initYourCommunication();//implement this
     *          passYourLVL0ControllersToTheRobotUnit();//implement this
     *          if(!validateAddedLVL0Controllers())
     *          {
     *              //report errors
     *              throw std::logic_error{"lvl0 controller setup invalid"};
     *          }
     *
     *          //at this point the RT thread becomes RT (calls before this may not be rt)
     *          IceUtil::Time currentStateTimestamp = TimeUtil::GetTime();
     *          IceUtil::Time lastStateTimestamp = TimeUtil::GetTime();
     *          getTheCurrentRobotState();//implement this
     *          while(isNotStopped())//implement this
     *          {
     *              const IceUtil::Time  startIteration = TimeUtil::GetTime();
     *              //switching controllers
     *              if(rtSwitchSetup())
     *              {
     *                  if(switchSetupError)
     *                  {
     *                      //report this error and exit
     *                      throw std::logic_error{"switching the controller setup failed"};
     *                  }
     *                  //validate the setup if you need to
     *                  //if you had to change something, call rtCommitActivatedControllers();
     *              }
     *              //run the lvl1 controllers
     *              rtRunLVL1Controllers(currentStateTimestamp, currentStateTimestamp - lastStateTimestamp);
     *              //validate targets
     *              for(std::size_t i = 0; i<rtGetActivatedLVL0Controllers().size(); ++i)
     *              {
     *                  if(!rtGetActivatedLVL0Controller(i)->isTargetVaild())
     *                  {
     *                      //handle this error!
     *                      //you probably should log this error to some error struct
     *                      rtDeactivateAssignedLVL1Controller(i);
     *                      rtCommitActivatedControllers();
     *                  }
     *              }
     *              //run lvl0
     *              rtRunLVL0Controllers();
     *
     *              //time keeping + communicating
     *              lastStateTimestamp = currentStateTimestamp;
     *              currentStateTimestamp = TimeUtil::GetTime();
     *              getTheCurrentRobotStateAndSendTheCurrentControllCommands();//implement this
     *              communicateWithYourPublisherThread();//implement this
     *              //maybe meassure how long a loop took and log this information to your error struct
     *              TimeUtil::Sleep(IceUtil::Time::microSeconds(1000)-(currentStateTimestamp-startIteration));
     *          }
     *      }
     *      catch(...)
     *      {
     *          emergencyStop();//implement this
     *          dumpDebuggingData();//implement this
     *          doSomeErrorHandling();//implement this
     *          shutEverythingGracefullyDown();//implement this
     *      }
     * }
     * \endcode
     *
     * \subsection RobotUnit-tasks-to-implement-pub Implementing the publisher thread
     *
     * \subsection RobotUnit-tasks-to-implement-com Implementing the communication between both threads
     * The communication between both threads has to be RT.
     * So you have to use atomics or TrippleBuffers (or any other non-blocking method)
     * Data you should communicate is:
     * \li Current state (joint modes/values)
     * \li Current goals
     * \li all error messages.
     *
     * \subsection RobotUnit-tasks-to-implement-lvl0 Adding LVL0 controllers
     * Add LVL0Controllers by calling
     * \code{.cpp}
     * addLVL0Controller(const std::string& jointName, LVL0Controller& lvl0Controller);
     * \endcode
     * If you add two controllers withe the same control mode for one joint, an std::invalid_argument exception is thrown.
     *
     * Each joint needs a LVL0Controller for the mode ControlModes::EmergencyStopMode.
     * This controller should activate the motor brakes.
     */
    class RobotUnit :
        virtual public Component,
        virtual public RobotUnitInterface,
        virtual public LVL1ControllerDataProviderInterface
    {
    public:
        using MutexType = std::recursive_mutex;
        using GuardType = std::lock_guard<MutexType>;

        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        //ice LVL1ControllerDataProviderInterface
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        virtual std::string getName() const override
        {
            return Component::getName();
        }
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        //ice interface
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        //controllers
        virtual Ice::StringSeq getControllersKnown(const Ice::Current& = GlobalIceCurrent) const override;
        virtual Ice::StringSeq getControllerNamesLoaded(const Ice::Current& = GlobalIceCurrent) const override;
        virtual Ice::StringSeq getControllerNamesRequested(const Ice::Current& = GlobalIceCurrent) const override;
        virtual Ice::StringSeq getControllerNamesActivated(const Ice::Current& = GlobalIceCurrent) const override;

        virtual StringLVL1ControllerPrxDictionary getControllersLoaded(const Ice::Current& = GlobalIceCurrent) const override;
        virtual StringLVL1ControllerPrxDictionary getControllersRequested(const Ice::Current& = GlobalIceCurrent) const override;
        virtual StringLVL1ControllerPrxDictionary getControllersActivated(const Ice::Current& = GlobalIceCurrent) const override;
        virtual JointNameToLVL1Dictionary getControllerJointAssignment(const Ice::Current& = GlobalIceCurrent) const override;
        //modes
        virtual JointNameToControlModeDictionary  getJointControlModesRequested(const Ice::Current& = GlobalIceCurrent) const override;
        virtual JointNameToControlModeDictionary  getJointControlModesActivated(const Ice::Current& = GlobalIceCurrent) const override;
        virtual JointNameToControlModesDictionary getJointControlModesSupported(const Ice::Current& = GlobalIceCurrent) const override;
        //loading and switching
        virtual void switchSetup(const Ice::StringSeq& controllerRequestedNames, const Ice::Current& = GlobalIceCurrent) override;
        virtual LVL1ControllerInterfacePrx loadController(const std::string& className, const std::string& instanceName, const LVL1ControllerConfigPtr& config, const Ice::Current& = GlobalIceCurrent) override;
        virtual bool loadLibFromPath(const std::string& path, const Ice::Current& = GlobalIceCurrent) override;
        virtual bool loadLibFromPackage(const std::string& package, const std::string& lib, const Ice::Current& = GlobalIceCurrent) override;

        //units
        virtual           KinematicUnitInterfacePrx getKinematicUnit(const Ice::Current& = GlobalIceCurrent) const = 0;
        virtual         ForceTorqueUnitInterfacePrx getForceTorqueUnit(const Ice::Current& = GlobalIceCurrent) const = 0;
        virtual InertialMeasurementUnitInterfacePrx getInertialMeasurementUnit(const Ice::Current& = GlobalIceCurrent) const = 0;
        virtual            PlatformUnitInterfacePrx getPlatformUnitInterface(const Ice::Current& = GlobalIceCurrent) const = 0;
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // used by lvl1 controllers
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        //units (nonrt but results are used in rt)
        virtual const ForceTorqueDataUnitInterface* getRTForceTorqueDataUnit() const = 0;
        virtual const      HapticDataUnitInterface* getRTHapticDataUnit() const = 0;
        virtual const         IMUDataUnitInterface* getRTIMUDataUnit() const = 0;
        virtual const   KinematicDataUnitInterface* getRTKinematicDataUnit() const = 0;
        virtual const    PlatformDataUnitInterface* getRTPlatformDataUnit() const = 0;
        //targets (nonrt but results are used in rt)
        virtual TargetBase* getJointTarget(const std::string& jointName, const std::string& controlMode) override;

    protected:
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // NonRT
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // setup the data structures
        void setJointNames(const std::vector<std::string>& names);

        // Component
        /// @see PropertyUser::createPropertyDefinitions()
        virtual armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;
        //get buffers
        /// @return The requested lvl0 controllers (none is null)
        std::vector<LVL0Controller*>& getRequestedLVL0Controllers()
        {
            return controllersRequested.getWriteBuffer().lvl0Controllers;
        }
        /// @return The requested lvl1 controllers (some may be null)
        std::vector<LVL1ControllerPtr  >& getRequestedLVL1Controllers()
        {
            return controllersRequested.getWriteBuffer().lvl1Controllers;
        }
        /// @return The requested lvl0 controllers (none is null)
        const std::vector<LVL0Controller*>& getRequestedLVL0Controllers() const
        {
            return controllersRequested.getWriteBuffer().lvl0Controllers;
        }
        /// @return The requested lvl1 controllers (some may be null)
        const std::vector<LVL1ControllerPtr  >& getRequestedLVL1Controllers() const
        {
            return controllersRequested.getWriteBuffer().lvl1Controllers;
        }
        /// @return The activated lvl0 controllers (none is null)
        const std::vector<LVL0Controller*>& getActivatedLVL0Controllers() const
        {
            return controllersActivated.getUpToDateReadBuffer().lvl0Controllers;
        }
        /// @return The activated lvl1 controllers (some may be null)
        const std::vector<LVL1ControllerPtr  >& getActivatedLVL1Controllers() const
        {
            return controllersActivated.getUpToDateReadBuffer().lvl1Controllers;
        }

        //checks
        /// @return Return whether the set of added LVL0Controllers is ok (e.g. if each joint has one EmergencyStop controller)
        virtual bool validateAddedLVL0Controllers() const;

        /**
         * @brief Add a LVL0Controller for a specified joint.
         * @param jointName The joint controlled by the controller
         * @param lvl0Controller The controller to add.
         * @throw If you add two controllers withe the same control mode for one joint, an std::invalid_argument exception is thrown.
         */
        void addLVL0Controller(const std::string& jointName, LVL0Controller& lvl0Controller);

        /// @return The LVL0Controller for given joint and control mode.
        const LVL0Controller* getLVL0Controller(const std::string& jointName, const std::string& controlMode) const;

        /// @return The LVL0Controller for given joint and control mode.
        LVL0Controller* getLVL0Controller(const std::string& jointName, const std::string& controlMode);

        LVL0Controller* rtGetEmergencyStopLVL0Controller(std::size_t index)
        {
            return lvl0ControllersEmergencyStop.at(index);
        }

        /// @return Whether a LVL0Controller for given joint and control mode exists.
        bool hasLVL0Controller(const std::string& jointName, const std::string& controlMode) const;

        /// @return Whether a LVL1Controller for this name is loaded.
        bool hasLVL1Controller(const std::string& name) const;

        static void rtSetLVL1ControllerActive(const LVL1ControllerPtr& lvl1, bool isActive = 1);

        //other
    private:
        bool areJointNamesSet() const;
        bool loadLibFromAbsolutePath(const std::string& path);

    protected:
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // RT
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // get buffers (rt)
        bool rtControllersWereSwitched() const
        {
            return controllersRequested.updateReadBuffer();
        }
        /// @return The requested lvl0 controllers (none is null)
        const std::vector<LVL0Controller*>& rtGetRequestedLVL0Controllers() const
        {
            return controllersRequested.getReadBuffer().lvl0Controllers;
        }
        /// @return The requested lvl1 controllers (some may be null)
        const std::vector<LVL1ControllerPtr  >& rtGetRequestedLVL1Controllers() const
        {
            return controllersRequested.getReadBuffer().lvl1Controllers;
        }
        /// @return The activated lvl0 controllers (none is null)
        std::vector<LVL0Controller*>& rtGetActivatedLVL0Controllers()
        {
            return controllersActivated.getWriteBuffer().lvl0Controllers;
        }
        /// @return The activated lvl1 controllers (some may be null)
        std::vector<LVL1ControllerPtr  >& rtGetActivatedLVL1Controllers()
        {
            return controllersActivated.getWriteBuffer().lvl1Controllers;
        }
        /// @return The activated lvl0 controllers (none is null)
        const std::vector<LVL0Controller*>& rtGetActivatedLVL0Controllers() const
        {
            return controllersActivated.getWriteBuffer().lvl0Controllers;
        }
        /// @return The activated lvl1 controllers (some may be null)
        const std::vector<LVL1ControllerPtr  >& rtGetActivatedLVL1Controllers() const
        {
            return controllersActivated.getWriteBuffer().lvl1Controllers;
        }
        //get controllers (rt)
        LVL0Controller* const& rtGetActivatedLVL0Controller(std::size_t index) const
        {
            return rtGetActivatedLVL0Controllers().at(index);
        }
        const LVL1ControllerPtr& rtGetActivatedLVL1Controller(std::size_t index) const
        {
            return rtGetActivatedLVL1Controllers().at(index);
        }
        LVL0Controller*& rtGetActivatedLVL0Controller(std::size_t index)
        {
            return rtGetActivatedLVL0Controllers().at(index);
        }
        LVL1ControllerPtr& rtGetActivatedLVL1Controller(std::size_t index)
        {
            return rtGetActivatedLVL1Controllers().at(index);
        }


        LVL0Controller* rtGetRequestedLVL0Controller(std::size_t index) const
        {
            return rtGetRequestedLVL0Controllers().at(index);
        }
        const LVL1ControllerPtr&   rtGetRequestedLVL1Controller(std::size_t index) const
        {
            return getRequestedLVL1Controllers().at(index);
        }

        void rtCommitActivatedControllers()
        {
            controllersActivated.commitWrite();
        }

        //switching (rt)
        bool rtSwitchSetup();
        void rtRunLVL1Controllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);
        void rtDeactivateAssignedLVL1Controller(std::size_t index);
        void rtRunLVL0Controllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);

        /**
         * @brief Hook for switching the lvl0 controller (this changes the controll mode)
         * @param index The index of the lvl0 controller
         * @param oldLVL0 The old lvl0 controller (will be deactivated)
         * @param newLVL0 The new lvl0 controller (will be activated)
         */
        virtual void rtSwitchLVL0Controller(std::size_t index, LVL0Controller* oldLVL0, LVL0Controller* newLVL0) = 0;


    protected:
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        // data
        // //////////////////////////////////////////////////////////////////////////////////////////////////////////////////// //
        mutable MutexType dataMutex;
        //data accessible in RT and nonRT
        std::vector<std::string> jointNames;
        std::unordered_map<std::string, std::size_t> jointNameIndices;
        /// @brief True if some error occured in switch LVL1Controllers
        std::atomic_bool switchSetupError {false};

    private:
        //controllers
        struct LVL0AndLVL1ControllerList
        {
            LVL0AndLVL1ControllerList(std::size_t size):
                lvl0Controllers {size, nullptr},
                            lvl1Controllers {size, nullptr}
            {}

            LVL0AndLVL1ControllerList() = default;
            LVL0AndLVL1ControllerList(const LVL0AndLVL1ControllerList&) = default;

            std::vector<LVL0Controller*> lvl0Controllers;
            std::vector<LVL1ControllerPtr  > lvl1Controllers;
        };
        /**
         * @brief Holds pointer to all lvl0 controllers. (index: [jointname][controlmode])
         * Is initialized by derived classes.
         * May not be accessed when the controll loop is running
         */
        std::map<std::string, std::map<std::string, LVL0Controller*>> lvl0Controllers;
        /**
         * @brief Holds all currently loaded LVL1 controllers (index: [instancename])
         * May not be accessed in rt.
         */
        std::map<std::string, LVL1ControllerPtr> lvl1Controllers;
        /// @brief LVL0Controllers for Emergency stop
        std::vector<LVL0Controller*> lvl0ControllersEmergencyStop;

        //used to communicate with rt
        /// @brief Controllers the RT thread is supposed to activate (direction nonRT -> RT) (some lvl1 controllers may be null)
        WriteBufferedTripleBuffer<LVL0AndLVL1ControllerList> controllersRequested;
        /// @brief Controllers the RT thread is currently running (direction RT -> nonRT) (some lvl1 controllers may be null)
        WriteBufferedTripleBuffer<LVL0AndLVL1ControllerList> controllersActivated;

        //other
        /// @brief Stores joints accessed via getJointTarget
        JointNameToControlModeDictionary jointsUsedByLVL1Controller;
        std::map<std::string, DynamicLibraryPtr> loadedLibs;
    };

    using RobotUnitPtr = IceInternal::Handle<RobotUnit>;
}

#endif
